def test_yaml_extension():

    txt = """
        name: "Model"

        symbols:
            controls: [alpha, beta]
            states: [hei, ho]

        equations:
            arbitrage: |

                β*(c[t+1]/c[t])^(-γ+)*r - 1
                β*(c[t+1]/c[t])^(-γ)*(r_2[t+1]-r_1[t+1])
        
        calibration:
            a: 0.1
            b: 10
    """

    import dolang  # monkey-patch yaml

    import yaml

    data = yaml.compose(txt)

    assert "name" in data
    assert "equation" not in data

    assert data["name"].value == "Model"

    assert [e.value for e in data["symbols"]["controls"]] == ["alpha", "beta"]

    assert [*data["symbols"].keys()] == ["controls", "states"]



import ast
import dolang
from dolang.grammar import str_expression as to_source
from dolang.symbolic import stringify, list_variables, list_symbols
from dolang.grammar import Tree


def test_parsing():
    from dolang.grammar import str_expression, sanitize
    from dolang.symbolic import parse_string

    e = parse_string("s + a(0) + b[t-1] + b[t] + b[t+1]")
    print(e.pretty())

    e = parse_string("chi*n^eta*c^sigma - w(1)")
    print(e.pretty())

    e = parse_string("chi*n^eta*c^sigma - w(1) | 0.01 <= n <= 1.0")
    print(e.pretty())

    e = parse_string("chi*n^eta*c^sigma - w(1) ⟂ 0.01 <= n <= 1.0")
    print(e.pretty())

    e = parse_string("i = exp(z)*k^alpha*n^(1-alpha) - (m)^(-1/sigma)")
    print(e.pretty())
    f = sanitize(e, variables=["m"])
    print(str_expression(f))

    s = "i = exp(z)*k^alpha*n^(1-alpha) - (m)^(-1/sigma)"
    e = parse_string(s)

    s = "i = exp(z)*k^alpha*n^(1-alpha) - (m)^(-1/sigma)"
    # vars = ['z', 'p', 'k', 'n', 'i', 'm', 'V', 'u', 'y', 'c', 'rk', 'w', 'y', 'c']
    # print("HI")
    # v = sanitize(s, variables=vars)
    # print(v)
    e = parse_string(s)
    print(str_expression(e))


def test_parsing_unicode():

    from dolang.symbolic import parse_string

    s = "αα"
    e = parse_string(s)
    print(e)


def test_expectation():
    from dolang.symbolic import parse_string

    s = "𝔼[ (x[t+1] / x[t]) ]"
    e = parse_string(s)
    print(e.pretty())
    from dolang.symbolic import str_expression

    print(str_expression(e))


def test_parse_string():
    from dolang.symbolic import parse_string

    e = parse_string("sin(a(1)+b+f(1)+f(4)+a[t+1])")
    assert isinstance(e, Tree)
    s = to_source(e)
    assert s == "sin(a[t+1] + b + f[t+1] + f[t+4] + a[t+1])"


def test_remove_timing():
    from dolang.symbolic import parse_string, remove_timing

    e = parse_string("sin(a(1)+b+f(1)+f(4)+a[t+1])")
    assert isinstance(e, Tree)
    s = to_source(e)
    rr = remove_timing(s)
    assert rr == "sin(a + b + f + f + a)"


def test_multiline():
    from dolang.symbolic import parse_string

    e = parse_string(
        "a ⟂ x <= y <= exp(z)\nb ⟂ x <= y <= z", start="complementarity_block"
    )
    assert len(e.children) == 2


def test_predicate():
    from dolang.symbolic import parse_string, str_expression

    e = parse_string("a[t] <= (x[t]+b)")
    print(e.pretty())
    print(str_expression(e))
    e = parse_string("∀t, a[t] <= (x[t]+b)")
    print(e.pretty())
    print(str_expression(e))


def test_subperiod():
    from dolang.symbolic import parse_string, str_expression

    e = parse_string("a[t$1] = a[t+1]")
    print(e.pretty())
    print(str_expression(e))
    e = parse_string("a[t$consumption] = a[t+1]")
    print(e.pretty())
    print(str_expression(e))
    from lark.lark import Lark


# def test_list_symbols_debug():
#     from dolang.symbolic import parse_string
#     e = parse_string('sin(a(1)+b+f(1)+f(4)+a(1)+a+a[t+1]+cos(0)')
#     l = ListSymbols(known_functions=['sin', 'f'])
#     l.visit(e)
#     # note that cos is recognized as variable
#     assert (l.variables == [(('a', 1), 4), (('a', 1), 22), (('cos', 0), 37)])
#     assert (l.constants == [('b', 9), ('a', 27)])
#     assert (l.functions == [('sin', 0), ('f', 11), ('f', 16)])
#     assert (l.problems == [['a', 0, 29, 'incorrect subscript']])


def test_list_symbols():
    from dolang.symbolic import parse_string

    e = parse_string("sin(a(1)+b+f(1)+f(4)+a(1))+cos(0)")
    ll = list_symbols(e)
    print(ll.variables)
    print(ll.parameters)
    # cos is recognized as a usual function
    assert ll.variables == [("a", 1), ("f", 1), ("f", 4)]
    assert ll.parameters == ["b"]


def test_list_variables():
    from dolang.symbolic import parse_string

    e = parse_string("sin(a(1)+b+f(1)+f(4)+sin(a)+k*cos(a(0)))")
    list_variables(e)
    assert list_variables(e) == [("a", 1), ("f", 1), ("f", 4), ("a", 0)]


def test_sanitize():

    from dolang.symbolic import sanitize, parse_string

    s = "sin(a(1)+b+a+f(-1)+f(4)+a(1))"

    expected = "sin(a[t+1] + b + a[t] + f[t-1] + f[t+4] + a[t+1])"
    assert sanitize(s, variables=["a", "f"]) == expected

    # # we also deal with = signs, and convert to python exponents
    # assert (sanitize("a(1) = a^3 + b") == "a(1) == (a) ** (3) + b")


def test_stringify():

    from dolang.symbolic import parse_string

    s = "sin(a(1) + b + a(0) + f(-1) + f(4) + a(1))"
    enes = stringify(s)
    assert enes == "sin(a__1_ + b_ + a__0_ + f_m1_ + f__4_ + a__1_)"


def test_time_shift():

    from dolang.symbolic import time_shift, stringify_parameter

    e = "sin(a(1) + b + a(0) + f(-1) + f(4) + a(1))"

    enes = stringify(time_shift(e, +1))
    assert (enes) == "sin(a__2_ + b_ + a__1_ + f__0_ + f__5_ + a__2_)"



import re
import ast


def eval_scalar(tree):
    try:
        if isinstance(tree, ast.Num):
            return tree.n
        elif isinstance(tree, ast.UnaryOp):
            if isinstance(tree.op, ast.USub):
                return -tree.operand.n
            elif isinstance(tree.op, ast.UAdd):
                return tree.operand.n
        else:
            raise Exception("Don't know how to do that.")
    except:
        raise Exception("Don't know how to do that.")


def parse(s):
    return ast.parse(s).body[0].value


class ExprVisitor(ast.NodeVisitor):
    def __init__(self, variables):
        self.variables = variables

    def visit_Call(self, call):
        name = call.func.id
        if name in self.variables:
            assert len(call.args) == 1
            n = eval_scalar(call.args[0])
            return self.visit_Variable((name, n))
        else:
            return self.visit_RCall(call)

    def visit_RCall(self, call):
        return self.generic_visit(call)

    def visit_Name(self, cname):
        name = cname.id
        if name in self.variables:
            return self.visit_Variable((name, 0))
        else:
            return self.visit_RName(cname)

    def visit_RName(self, name):
        return self.generic_visit(name)


class ExprTransformer(ast.NodeTransformer):
    def __init__(self, variables):
        self.variables = variables

    def visit_Call(self, call):
        name = call.func.id
        if name in self.variables:
            assert len(call.args) == 1
            n = eval_scalar(call.args[0])
            return self.visit_Variable((name, n))
        else:
            return self.generic_visit(call)

    def visit_Name(self, cname):
        name = cname.id
        if name in self.variables:
            return self.visit_Variable((name, 0))
        else:
            return self.generic_visit(cname)


class TimeShift(ExprVisitor):
    def __init__(self, variables, shift):
        self.variables = variables
        self.shift = shift

    def visit_Variable(self, tvar):
        name, t = tvar
        return parse("{}({})".format(name, t + self.shift))


class Apply(ExprVisitor):
    def __init__(self, variables, fun):
        self.variables = variables
        self.fun = fun

    def visit_Variable(self, tvar):
        return self.fun(tvar)


reg_rad = re.compile("([^_]+)")
reg_sep = re.compile("(&|_)")
reg_bar = re.compile("(.*)(bar|star)")

gl = [
    "alpha",
    "beta",
    "gamma",
    "delta",
    "eta",
    "epsilon",
    "iota",
    "kappa",
    "lambda",
    "mu",
    "nu",
    "rho",
    "pi",
    "sigma",
    "tau",
    "theta",
    "upsilon",
    "omega",
    "phi",
    "psi",
    "zeta",
    "xi",
    "chi",
    "Gamma",
    "Delta",
    "Lambda",
    "Sigma",
    "Theta",
    "Upsilon",
    "Omega",
    "Xi",
    "Pi",
    "Phi",
    "Psi",
]
gl_special = {"lam": "\\lambda"}
greek_letters = dict([(x, "\\" + x) for x in gl])
greek_letters.update(gl_special)


def greekify(expr):
    m = reg_bar.match(expr)
    if m:
        expr = m.group(1)
        suffix = m.group(2)
    else:
        suffix = None
    if expr in greek_letters:
        res = greek_letters[expr]
    else:
        res = expr
    if suffix == "bar":
        res = "\\overline{{{}}}".format(res)
    elif suffix == "star":
        res = "{}^{{\\star}}".format(res)
    return res


# greekify('zbar')


def split_name_into_parts(a):
    s = a.replace("__", "&")
    m = reg_rad.findall(a)
    rad = m[0]
    cont = m[1:]
    m = reg_sep.findall(s)
    exponents = []
    indices = []
    for i in range(len(cont)):
        if m[i] == "_":
            indices.append(cont[i])
        else:
            exponents.append(cont[i])
    return [rad, indices, exponents]


def name_to_latex(name, date=None):
    [rad, indices, exponents] = split_name_into_parts(name)
    rad = greekify(rad)
    indices = [greekify(r) for r in indices]
    exponents = [greekify(r) for r in exponents]
    up = "{" + str.join(",", exponents) + "}"
    down = "{" + str.join(",", indices) + "}"
    sup = "^{0}".format(up) if up != "{}" else ""
    sdown = "^{0}".format(down) if down != "{}" else ""

    up = "{" + str.join(",", exponents) + "}"

    if date is not None and date != "S":
        if date == 0:
            times = "t"
        elif date > 0:
            times = "t+" + str(date)
        elif date < 0:
            times = "t-" + str(-date)
        indices = indices + [times]
    # else:
    # raise(Exception('Time variable {0} has unknown date : {1}.'.format(name,date)))

    down = "{" + str.join(",", indices) + "}"

    if len(up) > 2 and len(down) > 2:
        resp = "{0}_{1}^{2}".format(rad, down, up)
    elif len(up) > 2:
        resp = "{0}^{1}".format(rad, up)
    elif len(down) > 2:
        resp = "{0}_{1}".format(rad, down)
    else:
        resp = "{0}".format(rad)

    if date == "S":
        resp = "\\overline{" + resp + "}"
    return resp


class LatexVisitor(ExprVisitor):
    def prec(self, n):
        return getattr(
            self, "prec_" + n.__class__.__name__, getattr(self, "generic_prec")
        )(n)

    def visit_Variable(self, tvar):
        return name_to_latex(tvar[0], tvar[1])

    def visit_RCall(self, n):
        func = self.visit(n.func)
        args = ", ".join(map(self.visit, n.args))
        if func == "sqrt":
            return "\\sqrt{%s}" % args
        else:
            return r"\operatorname{%s}\left(%s\right)" % (func, args)

    def prec_Call(self, n):
        return 1000

    def prec_RCall(self, n):
        return 1000

    def visit_RName(self, n):
        return name_to_latex(n.id)

    def prec_RName(self, n):
        return 1000

    def prec_Name(self, n):
        return 1000

    def visit_UnaryOp(self, n):
        if self.prec(n.op) > self.prec(n.operand):
            return r"%s \left(%s\right)" % (self.visit(n.op), self.visit(n.operand))
        else:
            return r"%s %s" % (self.visit(n.op), self.visit(n.operand))

    def prec_UnaryOp(self, n):
        return self.prec(n.op)

    def visit_BinOp(self, n):
        if self.prec(n.op) > self.prec(n.left):
            left = r"\left(%s\right)" % self.visit(n.left)
        else:
            left = self.visit(n.left)
        if self.prec(n.op) > self.prec(n.right):
            right = r"\left(%s\right)" % self.visit(n.right)
        elif self.prec(n.op) == self.prec(n.right) and isinstance(n.op, ast.Sub):
            right = r"\left(%s\right)" % self.visit(n.right)
        else:
            right = self.visit(n.right)
        if isinstance(n.op, ast.Div):
            return r"\frac{%s}{%s}" % (self.visit(n.left), self.visit(n.right))
        elif isinstance(n.op, ast.FloorDiv):
            return r"\left\lfloor\frac{%s}{%s}\right\rfloor" % (
                self.visit(n.left),
                self.visit(n.right),
            )
        elif isinstance(n.op, ast.Pow):
            return r"%s^{%s}" % (left, self.visit(n.right))
        else:
            return r"%s %s %s" % (left, self.visit(n.op), right)

    def prec_BinOp(self, n):
        return self.prec(n.op)

    def visit_Sub(self, n):
        return "-"

    def prec_Sub(self, n):
        return 300

    def visit_Add(self, n):
        return "+"

    def prec_Add(self, n):
        return 300

    def visit_Mult(self, n):
        return "\\;"

    def prec_Mult(self, n):
        return 400

    def visit_Mod(self, n):
        return "\\bmod"

    def prec_Mod(self, n):
        return 500

    def prec_Pow(self, n):
        return 700

    def prec_Div(self, n):
        return 400

    def prec_FloorDiv(self, n):
        return 400

    def visit_LShift(self, n):
        return "\\operatorname{shiftLeft}"

    def visit_RShift(self, n):
        return "\\operatorname{shiftRight}"

    def visit_BitOr(self, n):
        return "\\operatorname{or}"

    def visit_BitXor(self, n):
        return "\\operatorname{xor}"

    def visit_BitAnd(self, n):
        return "\\operatorname{and}"

    def visit_Invert(self, n):
        return "\\operatorname{invert}"

    def prec_Invert(self, n):
        return 800

    def visit_Not(self, n):
        return "\\neg"

    def prec_Not(self, n):
        return 800

    def visit_UAdd(self, n):
        return "+"

    def prec_UAdd(self, n):
        return 800

    def visit_USub(self, n):
        return "-"

    def prec_USub(self, n):
        return 800

    def visit_Num(self, n):
        return str(n.n)

    def prec_Num(self, n):
        return 1000

    def generic_visit(self, n):
        if isinstance(n, ast.AST):
            return r"" % (
                n.__class__.__name__,
                ", ".join(map(self.visit, [getattr(n, f) for f in n._fields])),
            )
        else:
            return str(n)

    def generic_prec(self, n):
        return 0


def expr2tex(variables, s):
    pt = ast.parse(s).body[0].value
    return LatexVisitor(variables).visit(pt)


def eq2tex(variables, s):

    expr = s.replace("==", "=").replace("=", "==")

    # TEMP FIX
    expr = expr.replace("[t]", "").replace("[t+1]", "(1)").replace("[t-1]", "(-1)")
    expr = expr.replace("^", "**")

    if "==" in expr:
        lhs, rhs = [expr2tex(variables, str.strip(e)) for e in str.split(expr, "==")]
        return "{} = {}".format(lhs, rhs)
    else:
        return expr2tex(variables, expr)


# ast.dump(ast.parse("a == b"))



# all symbolic functions take ast expression trees (not expressions) as input
# this one can be constructed as : ast.parse(s).body[0].value

import ast

from ast import UnaryOp, UAdd, USub, Name, Load, Call
from ast import NodeTransformer


class Compare:
    def __init__(self):
        self.d = {}

    def compare(self, A, B):
        if isinstance(A, ast.Name) and (A.id[0] == "_"):
            if A.id not in self.d:
                self.d[A.id] = B
                return True
            else:
                return self.compare(self.d[A.id], B)
        if not (A.__class__ == B.__class__):
            return False
        if isinstance(A, ast.Name):
            return A.id == B.id
        elif isinstance(A, ast.Call):
            if not self.compare(A.func, B.func):
                return False
            if not len(A.args) == len(B.args):
                return False
            for i in range(len(A.args)):
                if not self.compare(A.args[i], B.args[i]):
                    return False
            return True
        elif isinstance(A, ast.Num):
            return A.n == B.n
        elif isinstance(A, ast.Expr):
            return self.compare(A.value, B.value)
        elif isinstance(A, ast.Module):
            if not len(A.body) == len(B.body):
                return False
            for i in range(len(A.body)):
                if not self.compare(A.body[i], B.body[i]):
                    return False
            return True
        elif isinstance(A, ast.BinOp):
            if not isinstance(A.op, B.op.__class__):
                return False
            if not self.compare(A.left, B.left):
                return False
            if not self.compare(A.right, B.right):
                return False
            return True
        elif isinstance(A, ast.UnaryOp):
            if not isinstance(A.op, B.op.__class__):
                return False
            return self.compare(A.operand, B.operand)
        elif isinstance(A, ast.Subscript):
            if not self.compare(A.value, B.value):
                return False
            return self.compare(A.slice, B.slice)
        elif isinstance(A, ast.Index):
            return self.compare(A.value, B.value)
        elif isinstance(A, ast.Compare):
            if not self.compare(A.left, B.left):
                return False
            if not len(A.ops) == len(B.ops):
                return False
            for i in range(len(A.ops)):
                if not self.compare(A.ops[i], B.ops[i]):
                    return False
            if not len(A.comparators) == len(B.comparators):
                return False
            for i in range(len(A.comparators)):
                if not self.compare(A.comparators[i], B.comparators[i]):
                    return False
            return True
        elif isinstance(A, ast.In):
            return True
        elif isinstance(A, (ast.Eq, ast.LtE)):
            return True
        else:
            print(A.__class__)
            raise Exception("Not implemented")


def compare(a, b):
    comp = Compare()
    val = comp.compare(a, b)
    d = comp.d
    return val


def match(m, s):
    comp = Compare()
    val = comp.compare(m, s)
    d = comp.d
    if len(d) == 0:
        return val
    else:
        return d


known_functions = ["log", "exp", "sin", "cos"]

import ast


class ListNames(ast.NodeVisitor):
    def __init__(self):
        self.found = []

    def visit_Name(self, name):
        self.found.append(name.id)


def get_names(expr):
    ln = ListNames()
    ln.visit(expr)
    return [e for e in ln.found]


def eval_scalar(tree):
    try:
        if isinstance(tree, ast.Num):
            return tree.n
        elif isinstance(tree, ast.UnaryOp):
            if isinstance(tree.op, ast.USub):
                return -tree.operand.n
        else:
            raise Exception("Don't know how to do that.")
    except:
        raise Exception("Don't know how to do that.")


class ExpressionChecker(ast.NodeVisitor):
    def __init__(self, spec_variables, known_functions, known_constants):
        self.spec_variables = spec_variables
        self.known_functions = known_functions
        self.known_constants = known_constants
        self.functions = []
        self.variables = []
        self.problems = []

    def visit_Call(self, call):
        name = call.func.id
        colno = call.func.col_offset
        if name in self.spec_variables:
            try:
                assert len(call.args) == 1
                n = eval_scalar(call.args[0])
                allowed_timing = self.spec_variables[name]
                if allowed_timing is None or (n in allowed_timing):
                    self.variables.append((name, n, call.func.col_offset))
                else:
                    self.problems.append(
                        [name, n, colno, "incorrect_timing", allowed_timing]
                    )
            except Exception as e:
                print(e)
                self.problems.append([name, None, colno, "timing_error"])

        elif name in self.known_functions:
            self.functions.append((name, colno))
            for e in call.args:
                self.visit(e)
        else:
            self.problems.append([name, None, colno, "unknown_function"])

    def visit_Name(self, name):
        # colno = name.colno
        colno = name.col_offset
        n = 0
        name = name.id
        if name in self.spec_variables:
            allowed_timing = self.spec_variables[name]
            if (allowed_timing is None) or (n in allowed_timing):
                self.variables.append((name, n, colno))
            else:
                self.problems.append(
                    [name, n, colno, "incorrect_timing", allowed_timing]
                )
        elif name not in self.known_constants:
            self.problems.append([name, 0, colno, "unknown_variable"])


def check_expression(expr, spec_variables, known_functions=[]):

    from dolo.compiler.language import functions, constants

    func = list(functions.keys()) + known_functions

    ch = ExpressionChecker(spec_variables, func, constants)
    ch.visit(expr)
    return dict(functions=ch.functions, variables=ch.variables, problems=ch.problems)



from dolang.version import __version_info__, __version__
from dolang.yaml_tools import yaml

from dolang.grammar import parse_string
from dolang.grammar import (
    stringify,
    list_variables,
    list_symbols,
    time_shift,
    steady_state,
)
from dolang.grammar import str_expression



import math
import copy

functions = {
    "log": math.log,
    "exp": math.exp,
    "sin": math.sin,
    "cos": math.cos,
    "tan": math.tan,
    "atan": math.atan,
    "tanh": math.tanh,
    "atanh": math.atanh,
    "min": min,
    "max": max,
    "minimum": min,
    "maximum": max,
    "abs": abs,
}

constants = {
    "pi": math.pi,
}


# this is a stupid implementation
class Language:

    functions = functions
    constants = constants
    objects = []
    object_names = []
    yaml_tags = []
    signatures = []

    def append(self, obj):

        self.objects.append(obj)
        self.object_names.append(obj.__name__)
        self.yaml_tags.append("!" + obj.__name__)

        # try to get signature
        try:
            sig = obj.signature
        except:
            sig = None
        self.signatures.append(sig)

    def isvalid(self, name):

        if name[0] == "!":
            name = name[1:]

        return name in self.object_names

    def get_from_tag(self, tag):
        assert tag[0] == "!"
        i = self.object_names.index(tag[1:])
        obj = self.objects[i]
        return obj

    def get_signature(self, tag):
        i = self.object_names.index(tag[1:])
        obj = self.objects[i]
        try:
            sig = obj.signature
        except:
            sig = None
        return sig


LANG = Language()


def language_element(el):
    LANG.append(el)
    return el


class ModelError(Exception):
    pass


def eval_data(data: "yaml_structure", calibration={}):

    import warnings
    from yaml import MappingNode, SequenceNode, ScalarNode

    d = calibration.copy()
    d.update(functions)

    if isinstance(data, ScalarNode):

        val = data.value
        if isinstance(val, (float, int)):
            return val

        elif isinstance(val, str):
            # could be a string, could be an expression, could depend on other sections
            try:
                val = eval(val.replace("^", "**"), d)
            except Exception as e:
                warnings.warn(f"Impossible to evaluate expression: {val}")
            return val

        else:
            raise Exception("Unknown scalar node type.")

    elif isinstance(data, SequenceNode):

        tag = data.tag
        if tag != "tag:yaml.org,2002:seq" and not LANG.isvalid(tag):
            # unknown object type
            # lc = data.lc
            lc = data.start_mark
            msg = (
                f"Line {lc.line}, column {lc.column}.  Tag '{tag}' is not recognized.'"
            )
            raise ModelError(msg)

        # eval children
        children = [eval_data(ch, calibration) for ch in data]

        if tag == "tag:yaml.org,2002:seq":
            return children
        else:
            objclass = LANG.get_from_tag(tag)
            return objclass(*children)

    elif isinstance(data, MappingNode):

        if (data.tag != "tag:yaml.org,2002:map") and data.tag == "!Function":
            return eval_function(data, calibration)

        tag = data.tag
        if tag != "tag:yaml.org,2002:map" and not LANG.isvalid(tag):
            # unknown object type
            lc = data.start_mark
            # lc = data.lc
            msg = (
                f"Line {lc.line}, column {lc.column}.  Tag '{tag}' is not recognized.'"
            )
            raise ModelError(msg)

        if tag != "tag:yaml.org,2002:map":
            # check argument names (ignore types for now)
            objclass = LANG.get_from_tag(tag)
            signature = LANG.get_signature(tag)
            sigkeys = [*signature.keys()]
            for a in data.keys():
                ## TODO account for repeated greek arguments
                if (a not in sigkeys) and (
                    greek_translation.get(a, None) not in sigkeys
                ):
                    lc = data.start_mark
                    sigstring = str.join(
                        ", ", [f"{k}={str(v)}" for k, v in signature.items()]
                    )
                    msg = f"Line {lc.line}, column {lc.column}. Unexpected argument '{a}'. Expected: '{objclass.__name__}({sigstring})'"
                    raise ModelError(msg)
                else:
                    try:
                        sigkeys.remove(a)
                    except:
                        sigkeys.remove(greek_translation[a])
            # remove optional arguments
            for sig in sigkeys:
                sigval = signature[sig]
                if sigval is not None and ("Optional" in sigval):
                    sigkeys.remove(sig)

            if len(sigkeys) > 0:
                sigstring = str.join(
                    ", ", [f"{k}={str(v)}" for k, v in signature.items()]
                )
                lc = data.lc
                msg = f"Line {lc.line}, column {lc.col}. Missing argument(s) '{str.join(', ',sigkeys)}'. Expected: '{objclass.__name__}({sigstring})'"
                raise ModelError(msg)

        # eval children
        children = []
        for key, ch in data.items():
            evd = eval_data(ch, calibration=calibration)
            if tag != "tag:yaml.org,2002:map":
                exptype = signature.get(key, None)
                if exptype in ["Matrix", "Optional[Matrix]"]:
                    matfun = LANG.get_from_tag("!Matrix")
                    try:
                        evd = matfun(*evd)
                    except:
                        lc = data.lc
                        msg = f"Line {lc.line}, column {lc.col}. Argument '{key}' could not be converted to Matrix"
                        raise ModelError(msg)
                elif exptype in ["Vector", "Optional[Vector]"]:
                    vectfun = LANG.get_from_tag("!Vector")
                    try:
                        evd = vectfun(*evd)
                    except:
                        lc = data.lc
                        msg = f"Line {lc.line}, column {lc.col}. Argument '{key}' could not be converted to Vector"
                        raise ModelError(msg)
            children.append(evd)

        kwargs = {k: v for (k, v) in zip(data.keys(), children)}
        if tag == "tag:yaml.org,2002:map":
            return kwargs
        else:
            objclass = LANG.get_from_tag(tag)
            return objclass(**kwargs)

    else:
        raise Exception("Unknown data structure.")


def eval_function(data, calibration):
    args = tuple(data["arguments"])
    content = copy.deepcopy(data["value"])

    def fun(x):
        calib = calibration.copy()
        for i, a in enumerate(args):
            calib[a] = x[i]
        res = eval_data(content, calib)
        return res

    return fun


# GREEK TOLERANCE

greek_translation = {
    "Sigma": "Σ",
    "sigma": "σ",
    "rho": "ρ",
    "mu": "μ",
    "alpha": "α",
    "beta": "β",
}


def greekify_dict(arg):
    dd = dict()
    for k in arg:
        if k in greek_translation:
            key = greek_translation[k]
        else:
            key = k
        if key in dd:
            raise Exception(f"key {key} defined twice")
        dd[key] = arg[k]
    return dd


def greek_tolerance(fun):
    def f(*pargs, **args):
        nargs = greekify_dict(args)
        return fun(*pargs, **nargs)

    return f



import ast
from ast import (
    Assign,
    arg,
    FunctionDef,
    Module,
    Store,
    Subscript,
    Name,
    Load,
    Index,
    Num,
)
from ast import arguments as ast_arguments

from dolang.grammar import tree_to_ast
from dolang.factory import FlatFunctionFactory
from dolang.symbolic import parse_string
from dolang.codegen import to_source

import sys


def compile_factory(fff: FlatFunctionFactory):
    arguments = [*fff.arguments.keys()]
    funname = fff.funname

    unpacking = []
    for i, (arg_group_name, arg_group) in enumerate(fff.arguments.items()):
        for pos, sym in enumerate(arg_group):
            rhs = Subscript(
                value=Name(id=arg_group_name, ctx=Load()),
                slice=Index(Num(pos)),
                ctx=Load(),
            )
            val = Assign(targets=[Name(id=sym, ctx=Store())], value=rhs)
            unpacking.append(val)

    body = []

    for k, neq in fff.preamble.items():
        tree = parse_string(neq)
        val = tree_to_ast(tree).value
        line = Assign(targets=[Name(id=k, ctx=Store())], value=val)
        body.append(line)

    for n, (k, neq) in enumerate(fff.content.items()):
        tree = parse_string(neq)
        val = tree_to_ast(tree).value
        line = Assign(targets=[Name(id=k, ctx=Store())], value=val)
        body.append(line)
    #
    for n, (lhs, neq) in enumerate(fff.content.items()):
        line = Assign(
            targets=[
                Subscript(
                    value=Name(id="out", ctx=Load()), slice=Index(Num(n)), ctx=Store()
                )
            ],
            value=Name(id=lhs, ctx=Load()),
        )
        body.append(line)

    if sys.version_info >= (3, 8, 0):
        f = FunctionDef(
            name=funname,
            args=ast_arguments(
                posonlyargs=[],
                args=[arg(arg=a) for a in arguments] + [arg(arg="out")],
                vararg=None,
                kwarg=None,
                kwonlyargs=[],
                kw_defaults=[],
                defaults=[],
            ),
            body=unpacking + body,
            decorator_list=[],
        )
        mod = Module(body=[f], type_ignores=[])
    else:
        f = FunctionDef(
            name=funname,
            args=ast_arguments(
                args=[arg(arg=a) for a in arguments] + [arg(arg="out")],
                vararg=None,
                kwarg=None,
                kwonlyargs=[],
                kw_defaults=[],
                defaults=[],
            ),
            body=unpacking + body,
            decorator_list=[],
        )
        mod = Module(body=[f])

    mmod = ast.fix_missing_locations(mod)
    return mmod


def make_method_from_factory(
    fff: FlatFunctionFactory, vectorize=True, use_file=False, debug=False, compile=True
):
    mod = compile_factory(fff)

    if debug:
        from .codegen import to_source

        print(to_source(mod))

    if vectorize:
        coredims = [len(v) for k, v in fff.arguments.items()]
        signature = str.join(",", ["(n_{})".format(d) for d in coredims])
        n_out = len(fff.content)
        if n_out in coredims:
            signature += "->(n_{})".format(n_out)
            # ftylist = float64[:](*([float64[:]] * len(coredims)))
            fty = "void(*[float64[:]]*{})".format(len(coredims) + 1)
        else:
            signature += ",(n_{})".format(n_out)
            fty = "void(*[float64[:]]*{})".format(len(coredims) + 1)
    else:
        signature = None

    fun = eval_ast(mod)

    if compile:
        from numba import jit, guvectorize

        jfun = jit(fun, nopython=True)
        if vectorize:
            gufun = guvectorize([fty], signature, target="parallel", nopython=True)(fun)
            return jfun, gufun
    else:
        jfun = fun

    return jfun


def eval_ast(mod):
    context = {}

    import numpy

    context["inf"] = numpy.inf
    context["exp"] = numpy.exp
    context["log"] = numpy.log
    context["sin"] = numpy.sin
    context["cos"] = numpy.cos
    context["tan"] = numpy.tan
    context["tanh"] = numpy.tanh
    context["atan"] = numpy.arctan
    context["atanh"] = numpy.arctanh

    context["maximum"] = numpy.maximum
    context["minimum"] = numpy.minimum
    context["max"] = numpy.maximum
    context["min"] = numpy.minimum
    context["abs"] = numpy.abs

    name = mod.body[0].name
    mod = ast.fix_missing_locations(mod)
    code = compile(mod, "<string>", "exec")
    exec(code, context, context)
    fun = context[name]

    return fun



def test_compiler():
    from dolang.factory import FlatFunctionFactory
    from dolang.codegen import to_source

    fff = FlatFunctionFactory(
        dict(x="b+c*p1", y="(a+b)*c"),  # preamble
        dict(out_1="exp(y*p1)-a*p2", out_2="x+y"),
        dict(g1=["a", "b"], g2=["c"], g3=["p1", "p2"]),
        "testfun",
    )

    from dolang.function_compiler import make_method_from_factory

    fun = make_method_from_factory(fff)[0]

    import numpy as np

    out = np.array([0.3, 0.1])

    fun(np.array([0.1, 0.2]), np.array([10]), np.array([0.5, 0.3]), out)

    assert abs(out[0] - 4.45168907) < 1e-8
    assert abs(out[1] - 8.2) < 1e-8


def test_compiler_2():
    from dolang.factory import FlatFunctionFactory
    from dolang.codegen import to_source

    fff = FlatFunctionFactory(
        dict(x="b+c*p1", y="(a+b)*c"),  # preamble
        dict(out_1="exp(y*p1)-a*p2", out_2="x+y"),
        dict(g1=["a", "b"], g2=["c"], g3=["p1", "p2"]),
        "testfun",
    )

    from dolang.function_compiler import make_method_from_factory

    fun = make_method_from_factory(fff, compile=False, debug=True)

    import numpy as np

    out = np.array([0.3, 0.1])

    fun(np.array([0.1, 0.2]), np.array([10]), np.array([0.5, 0.3]), out)

    assert abs(out[0] - 4.45168907) < 1e-8
    assert abs(out[1] - 8.2) < 1e-8



# monkey-patch yaml so that yaml.compose becomes more usable.
import yaml
from yaml import MappingNode, SequenceNode, ScalarNode
from typing import Union

MappingNode.keys = lambda self: [e[0].value for e in self.value]
MappingNode.values = lambda self: [e[1] for e in self.value]
MappingNode.items = lambda self: iter([(e[0].value, e[1]) for e in self.value])
MappingNode.__getitem__ = lambda self, key: self.value[self.keys().index(key)][1]
MappingNode.__iter__ = lambda self: iter(self.keys())


def __mn__get__(self, key, default=None):
    if key in self.keys():
        return self[key]
    else:
        return default


MappingNode.get = __mn__get__


def __mn__setitem__(self, key: str, value: Union[str, float, int]):
    assert isinstance(key, str)
    if isinstance(value, str):
        new_value = ScalarNode(tag="tag:yaml.org,2002:str", value=value)
    elif isinstance(value, int):
        new_value = ScalarNode(tag="tag:yaml.org,2002:int", value=value)
    elif isinstance(value, float):
        new_value = ScalarNode(tag="tag:yaml.org,2002:float", value=value)
    else:
        raise Exception(f"Unknown type for {value}")
    if key not in self.keys():
        new_key = ScalarNode(tag="tag:yaml.org,2002:str", value="name")
        self.value.append((new_key, new_value))
    else:
        i = self.keys().index(key)
        k = self.value[i][0]
        self.value[i] = (k, new_value)


MappingNode.__setitem__ = __mn__setitem__

MappingNode.__getitem__ = lambda self, key: self.value[self.keys().index(key)][1]

MappingNode.__contains__ = lambda self, key: (key in self.keys())
MappingNode.__len__ = lambda self: len(self.value)


def __mn_update__(self, kw):
    for k, w in kw.items():
        self[k] = w


MappingNode.update = __mn_update__

SequenceNode.__getitem__ = lambda self, i: self.value[i]
SequenceNode.__len__ = lambda self: len(self.value)
SequenceNode.__iter__ = lambda self: iter(self.value)



from .grammar import *

import ast


def eval_scalar(tree):
    try:
        if isinstance(tree, ast.Num):
            return tree.n
        elif isinstance(tree, ast.UnaryOp):
            if isinstance(tree.op, ast.USub):
                return -tree.operand.n
        else:
            raise Exception("Don't know how to do that.")
    except:
        raise Exception("Don't know how to do that.")



from hashlib import new
from lark.exceptions import (
    LarkError,
    UnexpectedInput,
    ConfigurationError,
    UnexpectedCharacters,
)
from yaml import ScalarNode

import copy
import lark
from lark.tree import Tree
from lark.lexer import Token
from lark.visitors import Interpreter, Visitor, Transformer

from functools import wraps

from typing import Tuple, Dict, Set, Union, List

Expression = Union[Tree, Token]

from dataclasses import dataclass

## the following grammar represents equations
## it recognizes variables indexed as v[t] or v[t+k] where k is an integer
## for compatibility purposes, it also recognizes v(0) as v[t] unless
## v is a prespecified function.
## later, this compatibility feature will be turned off.

import os

DIR_PATH, this_filename = os.path.split(__file__)
DATA_PATH = os.path.join(DIR_PATH, "grammar.lark")

grammar_0 = open(DATA_PATH, "rt", encoding="utf-8").read()

# from .grammar_lark import grammar_0

from lark.lark import Lark

parser = Lark(
    grammar_0,
    start=[
        "start",
        "equation",
        "variable",
        "equation_block",
        "assignment_block",
        "complementarity_block",
    ],
)


def parse_string(text, start=None):

    if start is None:
        start = "start"

    if isinstance(text, ScalarNode):
        if text.tag != "tag:yaml.org,2002:str":
            #     raise Exception(f"Don't know how to parse node {text}")
            txt = text.value
        else:
            if text.start_mark is None:
                txt = text.value
            else:
                buffer = text.end_mark.buffer
                i1 = text.start_mark.pointer
                i2 = text.end_mark.pointer
                txt = buffer[i1:i2]
                if text.style in (">", "|"):
                    txt = txt[1:]

    else:
        txt = text

    try:
        return parser.parse(txt, start)

    except (UnexpectedInput, UnexpectedCharacters) as e:

        if isinstance(text, ScalarNode):

            sm = text.start_mark
            # em = text.end_mark
            if text.style not in (">", "|"):
                new_column = sm.column + e.column
                new_line = sm.line + e.line
            else:
                new_line = sm.line + e.line
                new_column = e.column
            # newargs = list(e.args)
            # newargs[0] = e.args[0].replace(f"line {e.line}", f"line {new_line}")
            # newargs[0] = newargs[0].replace(f"col {e.column}", f"col {new_column}")
            # e.args = tuple(newargs)
            e.line = new_line
            e.column = new_column

        raise e


# Prints a tree as a string
# WIP!!! (probably correct, but way too many parentheses)
class Printer(Interpreter):
    def add(self, tree):
        if len(tree.children) == 1:
            return "ERROR"
        a = self.visit(tree.children[0])
        b = self.visit(tree.children[1])
        return f"{a} + {b}"

    def sub(self, tree):
        if len(tree.children) == 1:
            return "ERROR"
        a = self.visit(tree.children[0])
        b = self.visit(tree.children[1])
        return f"{a} - ({b})"

    def variable(self, tree):
        name = tree.children[0].children[0].value

        # is there a subperiod?
        subperiod = [
            c for c in tree.children if (isinstance(c, Tree) and c.data == "subperiod")
        ]
        if len(subperiod) == 1:
            sp = subperiod[0].children[0].value
            return f"{name}[t${sp}]"

        try:
            time = int(tree.children[1].children[0].value)
        except:
            time = 0
        if time == 0:
            ds = "t"
        elif time > 0:
            ds = "t+" + str(time)
        elif time < 0:
            ds = "t-" + str(-time)
        return f"{name}[{ds}]"

    def equality(self, tree):
        a = self.visit(tree.children[0])
        b = self.visit(tree.children[1])
        return f"{a} = {b}"

    def double_complementarity(self, tree):
        a = self.visit(tree.children[0])
        b = self.visit(tree.children[1])
        return f"{a} ⟂ {b}"

    def double_inequality(self, tree):
        a = self.visit(tree.children[0])
        b = self.visit(tree.children[1])
        c = self.visit(tree.children[2])
        return f"{a} <= {b} <= {c}"

    def assignment(self, tree):
        a = self.visit(tree.children[0])
        b = self.visit(tree.children[1])
        return f"{a} = {b}"

    def symbol(self, tree):
        name = tree.children[0].value
        return name

    def mul(self, tree):
        a = self.visit(tree.children[0])
        b = self.visit(tree.children[1])
        return f"({a})*({b})"

    def div(self, tree):
        a = self.visit(tree.children[0])
        b = self.visit(tree.children[1])
        return f"({a})/({b})"

    def call(self, tree):
        funname = tree.children[0].value
        args = self.visit(tree.children[1])
        return f"{funname}({args})"

    def pow(self, tree):
        arg1 = self.visit(tree.children[0])
        arg2 = self.visit(tree.children[1])
        return f"({arg1})^({arg2})"

    def number(self, tree):
        return tree.children[0].value

    def signed_int(self, tree):
        return tree.children[0].value

    def neg(self, tree):
        a = self.visit(tree.children[0])
        return f"-({a})"

    def expectation(self, tree):
        a = self.visit(tree.children[0])
        return f"𝔼[ {a} ]"

    def inequality(self, tree):
        a = self.visit(tree.children[0])
        b = (tree.children[1]).value
        c = self.visit(tree.children[2])
        return f"{a} {b} {c}"

    def predicate(self, tree):
        if len(tree.children) == 1:
            return self.visit(tree.children[0])
        else:
            return "∀t, " + self.visit(tree.children[1])


def create_variable(name, time):
    return Tree(
        "variable",
        [
            Tree("name", [Token("NAME", name)]),
            Tree("date", [Token("NUMBER", str(time))]),
        ],
    )


## replaces v[t] by v[t+0] (I didn't find how to do it in the grammar)
## replaces v by v[t] when v identified as a variable
class Sanitizer(Transformer):
    def __init__(self, variables=[]):
        self.__variables__ = variables

    def symbol(self, *args):
        tok = args[0][0]
        val = tok.value
        if val in self.__variables__:
            return Tree(
                "variable",
                [
                    Tree("name", [Token("NAME", val)]),
                    Tree("date", [Token("NUMBER", "0")]),
                ],
            )
        else:
            return Tree("symbol", *args)

    def variable(self, *args):
        if len(args[0]) == 1:
            date = Tree("date", [Token("NUMBER", "0")])
            args = (args[0] + [date],)
        return Tree("variable", *args)


## removes timing (replace v[t], v[t-1] or v[t+1] by v)
class TimeRemover(Transformer):
    def variable(self, args):
        name = args[0].children[0].value
        return Tree("symbol", [Token("NAME", name)])


def stringify_variable(arg: Tuple[str, int]) -> str:
    s = arg[0]
    date = arg[1]
    if date == 0:
        return "{}__0_".format(s)
    elif date <= 0:
        return "{}_m{}_".format(s, str(-date))
    elif date >= 0:
        return "{}__{}_".format(s, str(date))


def stringify_parameter(p: str) -> str:
    return "{}_".format(p)


def stringify_symbol(arg) -> str:
    if isinstance(arg, str):
        return stringify_parameter(arg)
    elif isinstance(arg, tuple):
        if len(arg) == 2 and isinstance(arg[0], str) and isinstance(arg[1], int):
            return stringify_variable(arg)
    raise Exception("Unknown canonical form: {}".format(arg))


## replaces symbols and variables by their canonical string represantation:
## symbols s -> "s_"
## variable v(1) -> "v_p1_"
## variable v(-1) -> "v_m1_"
class Stringifier(Transformer):
    def symbol(self, children):

        name = children[0].value
        if name == "inf":
            s = name
        else:
            s = stringify_parameter(name)
        return Tree("symbol", [Token("NAME", s)])

    def variable(self, children):

        name = children[0].children[0].value
        if len(children) == 1:
            date = 0
        else:
            date = int(children[1].children[0].value)
        s = stringify_variable((name, date))
        return Tree("symbol", [Token("NAME", s)])


###
## if shift == 'S' replaces all v[t+k] by v[t] (is that reasonable ?)
## if shift is an integer, replaces v[t+k] by v[t+k+shift]
class TimeShifter(Transformer):
    def __init__(self, shift: int):
        self.shift = shift

    def variable(self, children):

        name = children[0].children[0].value
        try:
            date = int(children[1].children[0].value)
        except:
            date = 0
        if self.shift == "S":
            new_date = "0"
        else:
            new_date = str(date + self.shift)
        return Tree(
            "variable",
            [
                Tree("name", [Token("NAME", name)]),
                Tree("date", [Token("NUMBER", new_date)]),
            ],
        )


@dataclass
class SymbolList(dict):
    variables: List[Tuple[str, int]]
    parameters: List[str]
    functions: List[str]


## lists all variables in an  expression
## result field contains variables with timing, parameters, and functions used
class VariablesLister(Visitor):
    def __init__(self):
        self.result = SymbolList([], [], [])

    def variable(self, tree):
        children = tree.children

        name = children[0].children[0].value
        date = int(children[1].children[0].value)
        if (name, date) not in self.result.variables:
            self.result.variables.append((name, date))

    def symbol(self, tree):
        children = tree.children
        name = children[0].value
        if name not in self.result.parameters:
            self.result.parameters.append(name)

    def call(self, tree):
        children = tree.children
        name = children[0].value
        if name not in self.result.functions:
            self.result.functions.append(name)


## replaces symbols in an expression:
## NameSustituter({'h': parser.parse("e+1")}).visit(parser.parse("h+p"))  returns
## parser.parse("e+1+p")
class NameSubstituter(Transformer):

    # substitutes a symbol by an expression
    def __init__(self, substitutions: Dict[str, Expression]):
        self.substitutions = substitutions

    def symbol(self, children):
        name = children[0].value
        if name in self.substitutions:
            return self.substitutions[name]
        else:
            return Tree("symbol", children)


def subs(expr: str, substitutions):
    import copy

    s = dict()
    f = expr
    for k, v in substitutions.items():
        s[k] = parser.parse(v, start="start")
    ns = NameSubstituter(s)
    res = ns.transform(f)
    return str_expression(res)


# prints expression
def str_expression(expr: Expression) -> str:
    return Printer().visit(expr)


# decorator to define functions which operate
# either on Trees or on strings.
def expression_or_string(f):
    @wraps(f)
    def wrapper(*args, **kwds):
        if not isinstance(args[0], str):
            return f(*args, **kwds)
        else:
            a = parser.parse(args[0], start="start")
            nargs = tuple([a]) + args[1:]
            res = f(*nargs, **kwds)
            return str_expression(res)

    return wrapper


## these functions apply the vistors/transformers either on expressions or on strings
@expression_or_string
def stringify(expr: Expression):
    return Stringifier().transform(expr)


@expression_or_string
def time_shift(expr: Expression, n) -> Expression:

    return TimeShifter(n).transform(expr)


@expression_or_string
def steady_state(expr: Expression) -> Expression:
    return TimeShifter(shift="S").transform(expr)


@expression_or_string
def sanitize(expr: Expression, variables=[]):
    return Sanitizer(variables=variables).transform(expr)


@expression_or_string
def remove_timing(expr: Expression):
    return TimeRemover().transform(expr)


def list_symbols(expr: Union[Expression, str]) -> SymbolList:

    if isinstance(expr, str):
        expr = parser.parse(str, start="start")

    ll = VariablesLister()
    ll.visit(expr)

    return ll.result


def list_variables(expr: Union[Expression, str]):
    return list_symbols(expr).variables


def test():

    l = lark.Lark(grammar_0)

    s = "d + h + oij^2 - (1.0 + a) + sin(x) + v[t+1] + v[t] + x(0) + z[t]"
    p0 = l.parse(s)

    list_variables = ["h"]
    p = Sanitizer(list_variables).transform(p0)

    pp = time_shift(s, -2)
    pp_2 = time_shift(s, "S")

    print("Original:                 ", s)
    print("Steady-state:             ", steady_state(s))
    print("Sanitize (variables=[h]): ", sanitize(s, variables=["h"]))
    print("Stringify: ", stringify(s))

    print("Original Tree:")
    print(p0.pretty())


if __name__ == "__main__":
    test()


def tree_to_ast(tree):
    import ast
    import re

    __regex_eq__ = re.compile("([^=]*)=([^=]*)")

    txt = str_expression(tree)

    ss = txt.replace("^", "**")
    m = re.match(__regex_eq__, ss)
    if m:
        ss = str.join("==", m.groups())

    return ast.parse(ss).body[0]



import ast


class Compare:
    def __init__(self):
        self.d = {}

    def compare(self, A, B):
        if isinstance(A, ast.Name) and (A.id[0] == "_"):
            if A.id not in self.d:
                self.d[A.id] = B
                return True
            else:
                return self.compare(self.d[A.id], B)
        if not (A.__class__ == B.__class__):
            return False
        if isinstance(A, ast.Name):
            return A.id == B.id
        elif isinstance(A, ast.Call):
            if not self.compare(A.func, B.func):
                return False
            if not len(A.args) == len(B.args):
                return False
            for i in range(len(A.args)):
                if not self.compare(A.args[i], B.args[i]):
                    return False
            return True
        elif isinstance(A, ast.Num):
            return A.n == B.n
        elif isinstance(A, ast.Expr):
            return self.compare(A.value, B.value)
        elif isinstance(A, ast.Module):
            if not len(A.body) == len(B.body):
                return False
            for i in range(len(A.body)):
                if not self.compare(A.body[i], B.body[i]):
                    return False
            return True
        elif isinstance(A, ast.BinOp):
            if not isinstance(A.op, B.op.__class__):
                return False
            if not self.compare(A.left, B.left):
                return False
            if not self.compare(A.right, B.right):
                return False
            return True
        elif isinstance(A, ast.UnaryOp):
            if not isinstance(A.op, B.op.__class__):
                return False
            return self.compare(A.operand, B.operand)
        elif isinstance(A, ast.Subscript):
            if not self.compare(A.value, B.value):
                return False
            return self.compare(A.slice, B.slice)
        elif isinstance(A, ast.Index):
            return self.compare(A.value, B.value)
        elif isinstance(A, ast.Compare):
            if not self.compare(A.left, B.left):
                return False
            if not len(A.ops) == len(B.ops):
                return False
            for i in range(len(A.ops)):
                if not self.compare(A.ops[i], B.ops[i]):
                    return False
            if not len(A.comparators) == len(B.comparators):
                return False
            for i in range(len(A.comparators)):
                if not self.compare(A.comparators[i], B.comparators[i]):
                    return False
            return True
        elif isinstance(A, ast.In):
            return True
        elif isinstance(A, (ast.Eq, ast.LtE)):
            return True
        else:
            print(A.__class__)
            raise Exception("Not implemented")


def compare_strings(a, b):
    t1 = ast.parse(a)
    t2 = ast.parse(b)
    comp = Compare()
    val = comp.compare(t1, t2)
    d = comp.d
    return val


def match(m, s):
    if isinstance(m, str):
        m = ast.parse(m).body[0].value
    if isinstance(s, str):
        s = ast.parse(s).body[0].value
    comp = Compare()
    val = comp.compare(m, s)
    d = comp.d
    if len(d) == 0:
        return val
    else:
        return d


import ast


class ReplaceExpectation(ast.NodeTransformer):
    import ast

    def visit_Subscript(self, node):
        from ast import Expr, Call, Name, Load, Lambda, arguments, arg

        m = match("E[ _expr | _x in _set]", node)
        if m:
            x_s = m["_x"].id  # name of dummy vaqriable
            expr = m["_expr"]
            sset = m["_set"]
            res = Call(
                func=Name(id="Ex", ctx=Load()),
                args=[
                    Lambda(
                        args=arguments(
                            args=[arg(arg=x_s, annotation=None)],
                            vararg=None,
                            kwonlyargs=[],
                            kw_defaults=[],
                            kwarg=None,
                            defaults=[],
                        ),
                        body=expr,
                    ),
                    sset,
                ],
                keywords=[],
                starargs=None,
                kwargs=None,
            )
            return res
        m = match("Sum[ _expr | _x in _set]", node)
        if m:
            x_s = m["_x"].id  # name of dummy vaqriable
            expr = m["_expr"]
            sset = m["_set"]
            res = Call(
                func=Name(id="Sum", ctx=Load()),
                args=[
                    Lambda(
                        args=arguments(
                            args=[arg(arg=x_s, annotation=None)],
                            vararg=None,
                            kwonlyargs=[],
                            kw_defaults=[],
                            kwarg=None,
                            defaults=[],
                        ),
                        body=expr,
                    ),
                    sset,
                ],
                keywords=[],
                starargs=None,
                kwargs=None,
            )
            return res

        return node


def test_strings():

    tests = [
        ["b(_x)", "b(1)", True],
        ["_x(1)", "b(1)", True],
        ["_x(2)", "b(1)", False],
        ["a + b + c", "a + b + c", True],
        ["a + b + d", "a + b + c", False],
        ["a + b + _d", "a + b + c", True],
        ["a + _d + _d", "a + b + c", False],
        ["a + _d + _d", "a + c + c", True],
        ["a*_x+c*_x**2+d*_y", "a*x+c*x**2+d*y", True],
        ["a*_x+c*_x**2+d*_y", "a*x+c*y**2+d*z", False],
        ["-_a(_x) + _b(_x)", "-u(-1)+v(-1)", True],
        ["-_a(_x) + _b(_x)", "-u(11)+v(11)", True],
        ["-_a(_x) + _b(_x)", "-u(1)+v(2)", False],
        ["E[_x]", "E[1]", True],
        ["E[2]", "E[1]", False],
        ["E[1]", "X[1]", False],
        ["a in _x", "a in b", True],
        ["a in b", "a in c", False],
        ["a | _x", "a | c", True],
        ["a | b", "a | c", False],
        ["E[ _x | _y in _z]", "E[ a | b in c]", True],
        ["E[ a | b in c]", "E[ a | b in d]", False],
        ["E[ _x | _y in _z]", "E[ f(a) | b in c]", True],
        ["E[ _x | _y in _z]", "E[ f(a)^2 | b in c]", True],
        # ['E[ f(s) ]',  'E[ _x ]', True],
        ["_a == _b | _x <= _y <= _z", "a == b | x <= y <= z", True],
        ["_eq | _comp", "phi[s] + (-(f[s]-min_f))*1000    |  phi[s]", True],
        ["_eq | _comp", "phi[s] + (-(f[s]-min_f))*1000    | 0 <= phi[s]", True],
        ["_eq | 0 <= _x", "phi[s] + (-(f[s]-min_f))*1000  | 0 <= phi[s]", True],
        # ['_cond : _eq | _comp' ,  's>0 : a == 8 | x<=0<=y', True],
    ]
    replacement_rules = [
        ["E[ expr | x in S(s)]", "Esp(lambda x: expr, S(s))"],
        ["Sum[ expr | x in H(s)]", "Sum(lambda x: expr, H(s)"],
    ]

    for l in tests:
        print("{} : {} : {} : {}".format(compare_strings(l[0], l[1]), l[2], l[0], l[1]))
        # assert(compare_strings(l[0], l[1]) ==  l[2])


def fix_equation(eq):
    eqq = ReplaceExpectation().visit(eq)
    return eqq


#
#
# if __name__ == "__main__":
#     test_strings()
#     res = match("E[ _i | _v in _s]", "E[ y(1)*2 | y(1) in S(s)]")
#     res = match('_a == _b | _x <= _y <= _z', 'a == b | x <= y <= z')
#     if match("a_","b"):
#         print("yes")
#     if match("a","a"):
#         print("yes")
#     else:
#         print("no")



def test_symbols():
    from dolang.symbols import Symbol
    import yaml

    class bcolors:
        HEADER = "\033[95m"
        OKBLUE = "\033[94m"
        OKGREEN = "\033[92m"
        WARNING = "\033[93m"
        FAIL = "\033[91m"
        ENDC = "\033[0m"
        BOLD = "\033[1m"
        UNDERLINE = "\033[4m"

    txt = yaml.safe_load(open("tests/symbols.yaml", encoding="utf-8"))

    summary = []
    for short_sym, long_sym in txt["symbols"].items():

        sym = eval(long_sym)
        result = short_sym == (str(sym))
        summary.append([sym, long_sym, result])

        # try the other direction (match short symbol)

    for l in summary:
        ok = l[2]
        msg = (
            bcolors.OKBLUE + "✓" + bcolors.ENDC
            if ok
            else bcolors.FAIL + "✗" + bcolors.ENDC
        )
        print("{} | {:<50} -> {}".format(msg, str(l[1]), l[0]))



def test_syntax_errors():

    from dolang.grammar import parse_string
    import os
    import yaml

    DIR_PATH, this_filename = os.path.split(__file__)
    DATA_PATH = os.path.join(DIR_PATH, "syntax_errors.yaml")

    txt = open(DATA_PATH, "rt", encoding="utf-8").read()
    data = yaml.compose(txt)

    try:
        parse_string(data["equations"]["list"][1], start="equation")
    except Exception as e:
        assert e.line == 13
        assert e.column == 29

    try:
        parse_string(data["equations"]["block"], start="equation_block")
    except Exception as e:
        assert e.line == 17
        assert e.column == 29

    try:
        parse_string(data["equations"]["block2"], start="equation_block")
    except Exception as e:
        assert e.line == 25
        assert e.column == 29

    try:
        parse_string(data["equations"]["inline"][1], start="equation")
    except Exception as e:
        assert e.line == 27
        assert e.column == 26


test_syntax_errors()


def test_variable_definitions_errors():

    from dolang.grammar import parse_string
    from lark.exceptions import UnexpectedCharacters
    import os
    import yaml

    DIR_PATH, this_filename = os.path.split(__file__)
    DATA_PATH = os.path.join(DIR_PATH, "syntax_errors.yaml")

    txt = open(DATA_PATH, "rt", encoding="utf-8").read()
    data = yaml.compose(txt)

    for v in (data["definitions"]).value:
        k = v[0]
        v = v[1].value
        if v == "None":
            parse_string(k, start="variable")  # no problem
        else:
            line = int(v[0].value)
            column = int(v[1].value)
            try:
                parse_string(k, start="variable")  # no problem
            except UnexpectedCharacters as e:
                assert e.column == column
                assert e.line == line



import numpy as np


class NumericEval:
    def __init__(self, d, mini_language):

        self.d = d  # dictionary of substitutions
        for k, v in d.items():
            assert isinstance(k, str)

        self.__supported_functions___ = supported_functions
        self.__supported_functions_names___ = [
            fun.__name__ for fun in self.__supported_functions___
        ]

    def __call__(self, s):

        return self.eval(s)

    def eval(self, struct):

        t = struct.__class__.__name__
        method_name = "eval_" + t.lower()
        try:
            fun = getattr(self, method_name)
        except Exception as e:
            raise Exception("Unknown type {}".format(method_name))
        return fun(struct)

    def eval_float(self, s):

        return s

    def eval_int(self, s):

        return s

    def eval_str(self, s):

        # not safe
        return eval(s, self.d)

    def eval_list(self, l):

        return [self.eval(e) for e in l]

    def eval_dict(self, d):

        if len(d) == 1:
            k = list(d.keys())[0]
            if k in self.__supported_functions_names___:
                i = self.__supported_functions_names___.index(k)
                fun = self.__supported_functions___[i]

                args = d[k]
                if isinstance(args, dict):
                    eargs = self.eval(args)
                    res = fun(**eargs)
                elif isinstance(args, (list, tuple)):
                    eargs = self.eval(args)
                    res = fun(*eargs)
                else:
                    res = args
                return res

        return {k: self.eval(e) for k, e in d.items()}

    def eval_ordereddict(self, s):

        from collections import OrderedDict

        res = OrderedDict()
        for k in s.keys():
            v = self.eval(s[k])
            res[k] = v

        return res

    def eval_ndarray(self, array_in):
        import numpy

        array_out = numpy.zeros_like(array_in, dtype=float)
        for i in range(array_in.shape[0]):
            for j in range(array_in.shape[1]):
                array_out[i, j] = self.eval(array_in[i, j])
        return array_out

    def eval_nonetype(self, none):
        return None


###
### Mini language example
###


# def TensorMarkov(*args, **kwargs):
#     from dolo.numeric.discretization import tensor_markov
#     return tensor_markov(*args, **kwargs)

# def Normal(a):
#     return a

# def Approximation(**kwargs):
#     return {'approximation_space': kwargs}

# def rouwenhorst(rho=None, sigma=None, N=None):
#     from dolo.numeric.discretization import rouwenhorst
#     return rouwenhorst(rho,sigma,N)

# def AR1(rho, sigma, *pargs, **kwargs):
#     rho_array = np.array(rho, dtype=float)
#     sigma_array = np.atleast_2d( np.array(sigma, dtype=float) )
#     try:
#         assert(rho_array.ndim<=1)
#     except:
#         raise Exception("When discretizing a Vector AR1 process, the autocorrelation coefficient must be as scalar. Found: {}".format(rho_array))
#     try:
#         assert(sigma_array.shape[0] == sigma_array.shape[1])
#     except:
#         raise Exception("The covariance matrix for a Vector AR1 process must be square. Found: {}".format())
#     from dolo.numeric.discretization import multidimensional_discretization
#     [P,Q] = multidimensional_discretization(rho_array, sigma_array, *pargs, **kwargs)
#     return P,Q

# def MarkovChain(a,b):
#     return [a,b]

# supported_functions = [AR1, TensorMarkov, MarkovChain, Normal, Approximation]


# ###
# ###


# # Markov mini language

# if __name__ == '__main__':

#     import numpy

#     from collections import OrderedDict
#     options = OrderedDict(
#         smin= ['x',0.0],
#         smax= ['y','x'],
#         orders= [40,40],
#         markov=dict(a=12.0, b=0.9)
#     )


#     d = {'x': 0.01, 'y': 10.0}
#     print( NumericEval(d, supported_functions)(options) )


#         # define a markov chain in yaml
#     txt = '''
# tensor:

#     - rouwenhorst:
#         rho: 0.9
#         sigma: 0.4
#         N: 3

#     - markov:
#         a: 0.8
#         b: 1.2


#     '''
#     import yaml
#     s = yaml.safe_load(txt)

#     print(NumericEval(d, supported_functions)(s))



import copy
import ast
from ast import NodeVisitor

from .symbolic import eval_scalar


def parse(s):
    return ast.parse(s).body[0].value


def unique(seq):
    seen = set()
    for item in seq:
        if item not in seen:
            seen.add(item)
            yield item


def tshift(t, n):
    return (t[0], t[1] + n)


def get_atoms(string):

    from .symbolic import parse_string, list_symbols

    if not isinstance(string, str):
        return set([])
    expr = parse_string(string)
    syms = list_symbols(expr)

    return set(syms.parameters)


def triangular_solver(incidence, context=None):

    if context is None:

        context = dict()
        import math

        context["log"] = math.log
        context["exp"] = math.exp
        context["nan"] = float("nan")

    n = len(incidence)

    current = copy.deepcopy(incidence)

    solved = []
    max_steps = len(incidence)
    steps = 0

    while (steps < max_steps) and len(solved) < n:

        possibilities = [
            i for i in range(n) if (i not in solved) and (len(current[i]) == 0)
        ]

        for i in possibilities:
            for e in current:
                if i in e:
                    e.remove(i)
            solved.append(i)

        steps += 1

    if len(solved) < n:
        raise Exception("System is not triangular")
    return solved


def get_incidence(sdict):

    var_order = list(sdict.keys())
    var_set = set(var_order)
    expressions = [sdict[k] for k in var_order]
    incidence = []

    for i, eq in enumerate(expressions):

        atoms = get_atoms(eq)
        vars = var_set.intersection(atoms)
        inds = [var_order.index(v) for v in vars]
        incidence.append(inds)

    return incidence


def solve_triangular_system(system, values=None, context=None):

    system = dict(system)
    var_order = list(system.keys())
    ll = get_incidence(system)

    sol_order = triangular_solver(ll, context=context)
    d = copy.copy(values) if values else {}

    import math

    d["nan"] = float("nan")
    d["nanj"] = 0  # float('nan')*0j # this is actually (nan+nanj): super strange
    d["exp"] = math.exp
    d["log"] = math.log
    d["sin"] = math.sin
    d["cos"] = math.cos

    for i in sol_order:
        v = var_order[i]
        try:
            val = system[v]
            s = str(val).replace("^", "**")
            fv = eval(s, d, d)
            d[v] = eval(s, d, d)
        except Exception as e:  # in case d[v] is an int
            raise (e)

    resp = dict([(v, d[v]) for v in system.keys()])
    return resp


def get_deps(incidence, var, visited=None):

    # assert(var in incidence)
    assert isinstance(var, tuple) and len(var) == 2

    if visited is None:
        visited = (var,)
    elif var in visited:
        raise Exception("Non triangular system.")
    else:
        visited = visited + (var,)

    n = var[1]
    if abs(n) > 20:
        raise Exception(
            "Found variable with time {}. Something has probably gone wrong.".format(n)
        )

    deps = incidence[(var[0], 0)]
    if n != 0:
        deps = [tshift(e, n) for e in deps]

    resp = sum([get_deps(incidence, e, visited) for e in deps], [])

    resp.append(var)

    return resp


def test_triangular_solution():

    d = {"a": 0.1, "b": "a + 3", "c": "a + b"}
    sol = solve_triangular_system(d)
    for v in sol.values():
        assert isinstance(v, float)


if __name__ == "__main__":
    test_triangular_solution()



def test_patterns():

    import dolang
    from dolang.pattern import compare_strings
    import yaml

    class bcolors:
        HEADER = "\033[95m"
        OKBLUE = "\033[94m"
        OKGREEN = "\033[92m"
        WARNING = "\033[93m"
        FAIL = "\033[91m"
        ENDC = "\033[0m"
        BOLD = "\033[1m"
        UNDERLINE = "\033[4m"

    tests = yaml.safe_load(open("tests/patterns.yaml", encoding="utf-8"))

    pattern_tests = tests["patterns"]

    for l in pattern_tests:

        res = compare_strings(l[0], l[1])
        expected = l[2]
        ok = res == expected
        msg = (
            bcolors.OKBLUE + "✓" + bcolors.ENDC
            if ok
            else bcolors.FAIL + "✗" + bcolors.ENDC
        )
        print(
            "{} : {} : {} : ({} instead of {})".format(msg, l[0], l[1], res, expected)
        )
